from typing import Annotated, Callable, Dict, Iterable, List, Optional, Set

from pydantic import BaseModel, TypeAdapter
import pytest
import sortedcontainers as sc

import sortedcontainers_pydantic as sc_p


def test_get_constructor():
    sc_p._get_constructor(sc.SortedDict) == sc.SortedDict
    sc_p._get_constructor(sc.SortedDict[str, int]) == sc.SortedDict
    sc_p._get_constructor(sc.SortedList) == sc.SortedList
    sc_p._get_constructor(sc.SortedList[int]) == sc.SortedList
    sc_p._get_constructor(sc.SortedSet) == sc.SortedSet
    sc_p._get_constructor(sc.SortedSet[int]) == sc.SortedSet
    sc_p._get_constructor(sc_p.SortedDict) == sc.SortedDict
    sc_p._get_constructor(sc_p.SortedDict[str, int]) == sc.SortedDict
    sc_p._get_constructor(sc_p.SortedList) == sc.SortedList
    sc_p._get_constructor(sc_p.SortedList[int]) == sc.SortedList
    sc_p._get_constructor(sc_p.SortedSet) == sc.SortedSet
    sc_p._get_constructor(sc_p.SortedSet[int]) == sc.SortedSet

    # Test with a custom subclass
    class CustomSortedDict(sc.SortedDict): ...

    sc_p._get_constructor(CustomSortedDict) == CustomSortedDict
    sc_p._get_constructor(CustomSortedDict[str, int]) == CustomSortedDict

    class CustomSCPSortedDict(sc_p.SortedDict): ...

    sc_p._get_constructor(CustomSCPSortedDict) == CustomSCPSortedDict
    sc_p._get_constructor(CustomSCPSortedDict[str, int]) == CustomSCPSortedDict


class ReusableIterable:
    """Utility iterable class that can be used multiple times without exhausting."""

    def __init__(self, iterable_factory: Callable[[], Iterable]):
        self.iterable_factory = iterable_factory

    def __iter__(self):
        return iter(self.iterable_factory())


def test_sorted_dict():
    expected = sc.SortedDict({"c": 1, "a": 2, "b": 3})

    base_cases = [
        expected,
        {"c": 1, "a": 2, "b": 3},
        [("c", 1), ("a", 2), ("b", 3)],
        [("b", 3), ("c", 1), ("a", 2)],
        (("c", 1), ("a", 2), ("b", 3)),
        {("c", 1), ("a", 2), ("b", 3)},
        ReusableIterable(lambda: [("c", 1), ("a", 2), ("b", 3)]),
    ]
    coercion_cases = [
        sc.SortedDict([("c", 1.0), ("a", 2.0), ("b", 3.0)]),
        {"c": 1.0, "a": 2.0, "b": 3.0},
    ]

    for annotation, coerces in (
        # sortedcontainers_pydantic subclass
        (sc_p.SortedDict, False),
        (sc_p.SortedDict[str, int], True),
        # annotation
        (sc_p.AnnotatedSortedDict, False),
        (sc_p.AnnotatedSortedDict[str, int], True),
        # manual annotation
        (Annotated[sc.SortedDict, sc_p.SortedDictPydanticAnnotation], False),
        (Annotated[sc.SortedDict[str, int], sc_p.SortedDictPydanticAnnotation], True),
    ):
        ta = TypeAdapter(annotation)
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedDict)
            assert actual == expected
        assert ta.dump_json(expected).decode() == '{"a":2,"b":3,"c":1}'

        # Works inside Optional
        ta = TypeAdapter(Optional[annotation])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedDict)
            assert actual == expected
            assert ta.validate_python(None) is None
        assert ta.dump_json(expected).decode() == '{"a":2,"b":3,"c":1}'

        # Works inside list
        ta = TypeAdapter(list[annotation])
        for case in cases:
            case = [case]
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, list)
            assert isinstance(actual[0], sc.SortedDict)
            assert actual == [expected]
        assert ta.dump_json([expected]).decode() == '[{"a":2,"b":3,"c":1}]'

    assert TypeAdapter(sc_p.SortedDict).json_schema() == TypeAdapter(dict).json_schema()
    assert (
        TypeAdapter(sc_p.SortedDict[str, int]).json_schema()
        == TypeAdapter(Dict[str, int]).json_schema()
    )
    assert (
        TypeAdapter(Optional[sc_p.SortedDict]).json_schema()
        == TypeAdapter(Optional[dict]).json_schema()
    )
    assert (
        TypeAdapter(Optional[sc_p.SortedDict[str, int]]).json_schema()
        == TypeAdapter(Optional[Dict[str, int]]).json_schema()
    )

    class MyModel(BaseModel):
        sorted_dict: sc_p.SortedDict

    class MyModelWithArg(BaseModel):
        sorted_dict: sc_p.SortedDict[str, int]

    class MyModelWithOptional(BaseModel):
        sorted_dict: Optional[sc_p.SortedDict]

    class MyModelWithOptionalWithArg(BaseModel):
        sorted_dict: Optional[sc_p.SortedDict[str, int]]

    class MyModelWithAnnotated(BaseModel):
        sorted_dict: sc_p.AnnotatedSortedDict

    class MyModelWithAnnotatedWithArg(BaseModel):
        sorted_dict: sc_p.AnnotatedSortedDict[str, int]

    class MyModelManualAnnotation(BaseModel):
        sorted_dict: Annotated[sc.SortedDict, sc_p.SortedDictPydanticAnnotation]

    class MyModelManualAnnotationWithArg(BaseModel):
        sorted_dict: Annotated[sc.SortedDict[str, int], sc_p.SortedDictPydanticAnnotation]

    for model, coerces in (
        (MyModel, False),
        (MyModelWithArg, True),
        (MyModelWithOptional, False),
        (MyModelWithOptionalWithArg, True),
        (MyModelWithAnnotated, False),
        (MyModelWithAnnotatedWithArg, True),
        (MyModelManualAnnotation, False),
        (MyModelManualAnnotationWithArg, True),
    ):
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"model: {model}, case: {case}")
            instance = model(sorted_dict=case)
            assert isinstance(instance.sorted_dict, sc.SortedDict)
            assert instance.sorted_dict == expected
            assert instance.model_dump_json() == '{"sorted_dict":{"a":2,"b":3,"c":1}}'


def test_sorted_dict_with_key():
    expected = sc.SortedDict({"c": 1, "a": 2, "b": 3}, key=lambda x: -x)

    annotations = (
        # sortedcontainers_pydantic subclass
        sc_p.SortedDict,
        sc_p.SortedDict[str, int],
        # annotation
        sc_p.AnnotatedSortedDict,
        sc_p.AnnotatedSortedDict[str, int],
        # manual annotation
        Annotated[sc.SortedDict, sc_p.SortedDictPydanticAnnotation],
        Annotated[sc.SortedDict[str, int], sc_p.SortedDictPydanticAnnotation],
    )

    for annotation in annotations:
        ta = TypeAdapter(Annotated[annotation, sc_p.Key(lambda x: -x)])

        ta.validate_python({"c": 1, "a": 2, "b": 3}) == expected
        tuple(ta.validate_python([("c", 1), ("a", 2), ("b", 3)]).keys()) == ("c", "b", "a")
        ta.validate_python({"c": 1.0, "a": 2.0, "b": 3.0}) == expected

        # Wrap in Optional
        ta = TypeAdapter(Optional[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python({"c": 1, "a": 2, "b": 3}) == expected
        tuple(ta.validate_python([("c", 1), ("a", 2), ("b", 3)]).keys()) == ("c", "b", "a")
        assert ta.validate_python(None) is None

        # Wrap in list
        ta = TypeAdapter(list[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python([{"c": 1, "a": 2, "b": 3}]) == [expected]


def test_sorted_list():
    expected = sc.SortedList([3, 1, 2])

    base_cases = [
        expected,
        [1, 2, 3],
        [3, 2, 1],
        [3, 1, 2],
        (3, 1, 2),
        {3, 1, 2},
        range(3, 0, -1),
        [2, 3, 1],
    ]
    coercion_cases = [
        sc.SortedList([3.0, 1.0, 2.0]),
        [2.0, 3.0, 1.0],
    ]

    for annotation, coerces in (
        # sortedcontainers_pydantic subclasses
        (sc_p.SortedList, False),
        (sc_p.SortedList[int], True),
        # annotations
        (sc_p.AnnotatedSortedList, False),
        (sc_p.AnnotatedSortedList[int], True),
        # manual annotation
        (Annotated[sc.SortedList, sc_p.SortedListPydanticAnnotation], False),
        (Annotated[sc.SortedList[int], sc_p.SortedListPydanticAnnotation], True),
    ):
        ta = TypeAdapter(annotation)
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedList)
            assert actual == expected
        assert ta.dump_json(expected).decode() == "[1,2,3]"

        # Works inside Optional
        ta = TypeAdapter(Optional[annotation])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedList)
            assert actual == expected
            assert ta.validate_python(None) is None
        assert ta.dump_json(expected).decode() == "[1,2,3]"

        # Works inside list
        ta = TypeAdapter(list[annotation])
        for case in cases:
            case = [case]
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, list)
            assert isinstance(actual[0], sc.SortedList)
            assert actual == [expected]
        assert ta.dump_json([expected]).decode() == "[[1,2,3]]"

    assert TypeAdapter(sc_p.SortedList).json_schema() == TypeAdapter(list).json_schema()
    assert TypeAdapter(sc_p.SortedList[int]).json_schema() == TypeAdapter(List[int]).json_schema()
    assert (
        TypeAdapter(Optional[sc_p.SortedList]).json_schema()
        == TypeAdapter(Optional[list]).json_schema()
    )
    assert (
        TypeAdapter(Optional[sc_p.SortedList[int]]).json_schema()
        == TypeAdapter(Optional[List[int]]).json_schema()
    )

    class MyModel(BaseModel):
        sorted_list: sc_p.SortedList

    class MyModelWithArg(BaseModel):
        sorted_list: sc_p.SortedList[int]

    class MyModelWithAnnotated(BaseModel):
        sorted_list: sc_p.AnnotatedSortedList

    class MyModelWithAnnotatedWithArg(BaseModel):
        sorted_list: sc_p.AnnotatedSortedList[int]

    class MyModelManualAnnotation(BaseModel):
        sorted_list: Annotated[sc.SortedList, sc_p.SortedListPydanticAnnotation]

    class MyModelManualAnnotationWithArg(BaseModel):
        sorted_list: Annotated[sc.SortedList[int], sc_p.SortedListPydanticAnnotation]

    for model, coerces in (
        (MyModel, False),
        (MyModelWithArg, True),
        (MyModelWithAnnotated, False),
        (MyModelWithAnnotatedWithArg, True),
        (MyModelManualAnnotation, False),
        (MyModelManualAnnotationWithArg, True),
    ):
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"model: {model}, case: {case}")
            instance = model(sorted_list=case)
            assert isinstance(instance.sorted_list, sc.SortedList)
            assert instance.sorted_list == expected
            assert instance.model_dump_json() == '{"sorted_list":[1,2,3]}'


def test_sorted_list_with_key():
    expected = sc.SortedList([3, 1, 2], key=lambda x: -x)

    annotations = (
        # sortedcontainers_pydantic subclass
        sc_p.SortedList,
        sc_p.SortedList[int],
        # annotation
        sc_p.AnnotatedSortedList,
        sc_p.AnnotatedSortedList[int],
        # manual annotation
        Annotated[sc.SortedList, sc_p.SortedListPydanticAnnotation],
        Annotated[sc.SortedList[int], sc_p.SortedListPydanticAnnotation],
    )

    for annotation in annotations:
        ta = TypeAdapter(Annotated[annotation, sc_p.Key(lambda x: -x)])

        ta.validate_python([3, 1, 2]) == expected
        tuple(ta.validate_python([3, 1, 2])) == (3, 2, 1)

        # Wrap in Optional
        ta = TypeAdapter(Optional[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python([3, 1, 2]) == expected
        assert ta.validate_python(None) is None

        # Wrap in list
        ta = TypeAdapter(list[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python([[3, 1, 2]]) == [expected]


def test_sorted_set():
    expected = sc.SortedSet([3, 1, 2])

    base_cases = [
        expected,
        [3, 1, 2],
        [1, 2, 3],
        [3, 2, 1],
        (3, 1, 2),
        {3, 1, 2},
        range(3, 0, -1),
        [2, 3, 1],
    ]
    coercion_cases = [
        sc.SortedSet([3.0, 1.0, 2.0]),
        [2.0, 3.0, 1.0],
    ]

    for annotation, coerces in (
        # sortedcontainers_pydantic subclass
        (sc_p.SortedSet, False),
        (sc_p.SortedSet[int], True),
        # annotation
        (sc_p.AnnotatedSortedSet, False),
        (sc_p.AnnotatedSortedSet[int], True),
        # manual annotation
        (Annotated[sc.SortedSet, sc_p.SortedSetPydanticAnnotation], False),
        (Annotated[sc.SortedSet[int], sc_p.SortedSetPydanticAnnotation], True),
    ):
        ta = TypeAdapter(annotation)
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedSet)
            assert actual == expected
        assert ta.dump_json(expected).decode() == "[1,2,3]"

        # Works inside Optional
        ta = TypeAdapter(Optional[annotation])
        for case in cases:
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, sc.SortedSet)
            assert actual == expected
            assert ta.validate_python(None) is None
        assert ta.dump_json(expected).decode() == "[1,2,3]"

        # Works inside list
        ta = TypeAdapter(list[annotation])
        for case in cases:
            case = [case]
            print(f"annotation: {annotation}, case: {case}")
            actual = ta.validate_python(case)
            assert isinstance(actual, list)
            assert isinstance(actual[0], sc.SortedSet)
            assert actual == [expected]
        assert ta.dump_json([expected]).decode() == "[[1,2,3]]"

    assert TypeAdapter(sc_p.SortedSet).json_schema() == TypeAdapter(set).json_schema()
    assert TypeAdapter(sc_p.SortedSet[int]).json_schema() == TypeAdapter(Set[int]).json_schema()
    assert (
        TypeAdapter(Optional[sc_p.SortedSet]).json_schema()
        == TypeAdapter(Optional[set]).json_schema()
    )
    assert (
        TypeAdapter(Optional[sc_p.SortedSet[int]]).json_schema()
        == TypeAdapter(Optional[Set[int]]).json_schema()
    )

    class MyModel(BaseModel):
        sorted_set: sc_p.SortedSet

    class MyModelWithArg(BaseModel):
        sorted_set: sc_p.SortedSet[int]

    class MyModelWithAnnotated(BaseModel):
        sorted_set: sc_p.AnnotatedSortedSet

    class MyModelWithAnnotatedWithArg(BaseModel):
        sorted_set: sc_p.AnnotatedSortedSet[int]

    class MyModelManualAnnotation(BaseModel):
        sorted_set: Annotated[sc.SortedSet, sc_p.SortedSetPydanticAnnotation]

    class MyModelManualAnnotationWithArg(BaseModel):
        sorted_set: Annotated[sc.SortedSet[int], sc_p.SortedSetPydanticAnnotation]

    for model, coerces in (
        (MyModel, False),
        (MyModelWithArg, True),
        (MyModelWithAnnotated, False),
        (MyModelWithAnnotatedWithArg, True),
        (MyModelManualAnnotation, False),
        (MyModelManualAnnotationWithArg, True),
    ):
        cases = base_cases + (coercion_cases if coerces else [])
        for case in cases:
            print(f"model: {model}, case: {case}")
            instance = model(sorted_set=case)
            assert isinstance(instance.sorted_set, sc.SortedSet)
            assert instance.sorted_set == expected
            assert instance.model_dump_json() == '{"sorted_set":[1,2,3]}'


def test_sorted_set_with_key():
    expected = sc.SortedSet([3, 1, 2], key=lambda x: -x)

    annotations = (
        # sortedcontainers_pydantic subclass
        sc_p.SortedSet,
        sc_p.SortedSet[int],
        # annotation
        sc_p.AnnotatedSortedSet,
        sc_p.AnnotatedSortedSet[int],
        # manual annotation
        Annotated[sc.SortedSet, sc_p.SortedSetPydanticAnnotation],
        Annotated[sc.SortedSet[int], sc_p.SortedSetPydanticAnnotation],
    )

    for annotation in annotations:
        ta = TypeAdapter(Annotated[annotation, sc_p.Key(lambda x: -x)])

        ta.validate_python([3, 1, 2]) == expected
        tuple(ta.validate_python([3, 1, 2])) == (3, 2, 1)

        # Wrap in Optional
        ta = TypeAdapter(Optional[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python([3, 1, 2]) == expected
        assert ta.validate_python(None) is None

        # Wrap in list
        ta = TypeAdapter(list[Annotated[annotation, sc_p.Key(lambda x: -x)]])
        ta.validate_python([[3, 1, 2]]) == [expected]


def test_annotation_with_bad_source_type():
    for annotation in (
        sc_p.SortedDictPydanticAnnotation,
        sc_p.SortedListPydanticAnnotation,
        sc_p.SortedSetPydanticAnnotation,
    ):
        with pytest.raises(sc_p.UnsupportedSourceTypeError):
            TypeAdapter(Annotated[list, annotation])


def test_key_with_bad_source_type():
    with pytest.raises(sc_p.UnsupportedSourceTypeError):
        TypeAdapter(Annotated[dict, sc_p.Key(lambda x: x)])



from typing import TYPE_CHECKING, Annotated

from pydantic import BaseModel

from sortedcontainers_pydantic import (
    AnnotatedSortedDict,
    AnnotatedSortedList,
    AnnotatedSortedSet,
    Key,
    SortedDict,
    SortedList,
    SortedSet,
)


class MyModel(BaseModel):
    # sortedcontainers_pydantic classes
    dct: SortedDict[str, int] = SortedDict()
    lst: SortedList[int] = SortedList()
    st: SortedSet[int] = SortedSet()

    # sortedcontainers_pydantic classes with key
    dct_key: Annotated[SortedDict[str, int], Key(lambda x: -x)] = SortedDict()
    lst_key: Annotated[SortedList[int], Key(lambda x: -x)] = SortedList()
    st_key: Annotated[SortedSet[int], Key(lambda x: -x)] = SortedSet()

    # annotations
    dct_annot: AnnotatedSortedDict[str, int] = SortedDict()
    lst_annot: AnnotatedSortedList[int] = SortedList()
    st_annot: AnnotatedSortedSet[int] = SortedSet()

    # annotations with key
    dct_annot_key: Annotated[AnnotatedSortedDict[str, int], Key(lambda x: -x)] = SortedDict()
    lst_annot_key: Annotated[AnnotatedSortedList[int], Key(lambda x: -x)] = SortedList()
    st_annot_key: Annotated[AnnotatedSortedSet[int], Key(lambda x: -x)] = SortedSet()


if TYPE_CHECKING:
    m = MyModel()

    reveal_type(m.dct)  # noqa: F821
    reveal_type(m.lst)  # noqa: F821
    reveal_type(m.st)  # noqa: F821

    reveal_type(m.dct_key)  # noqa: F821
    reveal_type(m.lst_key)  # noqa: F821
    reveal_type(m.st_key)  # noqa: F821

    reveal_type(m.dct_annot)  # noqa: F821
    reveal_type(m.lst_annot)  # noqa: F821
    reveal_type(m.st_annot)  # noqa: F821

    reveal_type(m.dct_annot_key)  # noqa: F821
    reveal_type(m.lst_annot_key)  # noqa: F821
    reveal_type(m.st_annot_key)  # noqa: F821

MyModel(dct={"c": 1, "a": 2, "b": 3})
MyModel(dct=[("c", 1), ("a", 2), ("b", 3)])
MyModel(dct=(pair for pair in [("c", 1), ("a", 2), ("b", 3)]))

MyModel(lst=[3, 1, 2])
MyModel(lst=(3, 1, 2))
MyModel(lst={3, 1, 2})
MyModel(lst=(i for i in [3, 1, 2]))

MyModel(st=[3, 1, 2])
MyModel(st=(3, 1, 2))
MyModel(st={3, 1, 2})
MyModel(st=(i for i in [3, 1, 2]))



from dataclasses import dataclass
from functools import partial
import importlib.metadata
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Hashable,
    Iterable,
    Mapping,
    Set,
    Tuple,
    TypeVar,
    get_args,
    get_origin,
)

from pydantic import (
    GetCoreSchemaHandler,
)
from pydantic_core import core_schema
import sortedcontainers

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison


__version__ = importlib.metadata.version("sortedcontainers-pydantic")

__all__ = [
    "SortedDict",
    "SortedList",
    "SortedSet",
    "SortedKeyList",
    "SortedDictPydanticAnnotation",
    "SortedListPydanticAnnotation",
    "SortedSetPydanticAnnotation",
    "AnnotatedSortedDict",
    "AnnotatedSortedList",
    "AnnotatedSortedSet",
    "Key",
    "UnsupportedSourceTypeError",
]

_KT = TypeVar("_KT", bound=Hashable)  # Key type.
_VT = TypeVar("_VT")  # Value type.
_T = TypeVar("_T")
_OrderableT = TypeVar("_OrderableT", bound="SupportsRichComparison")
_HashableT = TypeVar("_HashableT", bound=Hashable)


class _UnsupportedSourceTypeError(Exception):
    def __init__(self, parsed: Any):
        self.parsed = parsed


class UnsupportedSourceTypeError(TypeError):
    pass


def _get_constructor(tp: Any) -> Any:
    """Get the relevant class constructor for the given type annotation, e.g., SortedList from
    SortedList[int] or appropriate subclass.
    """
    # Parse the annotation to get the relevant source type, e.g., SortedList form SortedList[int]
    origin = get_origin(tp)
    if origin is None:
        parsed = tp
    else:
        parsed = origin

    bases = set((parsed,) + getattr(parsed, "__bases__", tuple()))
    if bases & {
        # Subclasses of sortedcontainers_pydantic
        SortedList,
        SortedDict,
        SortedSet,
        # Subclasses of sortedcontainers
        sortedcontainers.SortedList,
        sortedcontainers.SortedDict,
        sortedcontainers.SortedSet,
    }:
        return parsed
    else:
        raise _UnsupportedSourceTypeError(parsed)


class SortedDictPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Returns pydantic_core.CoreSchema that defines how Pydantic should validate and
        serialize this class.

        - Validating from JSON: Validate as an iterable and pass to SortedList
          constructor
        - Validating from Python:
            - If it's already a SortedList, do nothing
            - If it's an iterable, pass to SortedList constructor
        - Serialization: Convert to a list
        """
        if cls is SortedDictPydanticAnnotation:
            # Used as annotation, i.e., Annotated[..., SortedDictPydanticAnnotation]
            try:
                cls = _get_constructor(source_type)
            except _UnsupportedSourceTypeError as e:
                msg = (
                    "Expected subclass of sortedcontainers.SortedDict, "
                    f"got '{e.parsed}' parsed from annotation '{source_type}'."
                )
                raise UnsupportedSourceTypeError(msg) from e

        # Schema for when the input is already an instance of this class
        instance_schema = core_schema.is_instance_schema(cls)

        # Get schema for Iterable type based on source type has arguments
        args = get_args(source_type)
        if args:
            mapping_t_schema = handler.generate_schema(Mapping[args[0], args[1]])  # type: ignore[valid-type]
            iterable_of_pairs_t_schema = handler.generate_schema(Iterable[Tuple[args[0], args[1]]])  # type: ignore[valid-type]
        else:
            mapping_t_schema = handler.generate_schema(Mapping)
            iterable_of_pairs_t_schema = handler.generate_schema(Iterable[Tuple[Any, Any]])

        # Schema for when the input is a mapping
        from_mapping_schema = core_schema.no_info_after_validator_function(
            function=cls, schema=mapping_t_schema
        )

        # Schema for when the input is an iterable of pairs
        from_iterable_of_pairs_schema = core_schema.no_info_after_validator_function(
            function=cls, schema=iterable_of_pairs_t_schema
        )

        # Union the schemas
        # Only include instance_schema if there are no type arguments
        # Otherwise an existing instance with wrong argument types won't be coerced
        if args:
            python_schema = core_schema.union_schema(
                [from_mapping_schema, from_iterable_of_pairs_schema]
            )
        else:
            python_schema = core_schema.union_schema(
                [instance_schema, from_mapping_schema, from_iterable_of_pairs_schema]
            )

        as_dict_serializer = core_schema.plain_serializer_function_ser_schema(dict)

        return core_schema.json_or_python_schema(
            json_schema=from_mapping_schema,
            python_schema=python_schema,
            serialization=as_dict_serializer,
        )


class SortedDict(sortedcontainers.SortedDict[_KT, _VT], SortedDictPydanticAnnotation):
    pass


AnnotatedSortedDict = Annotated[
    sortedcontainers.SortedDict[_KT, _VT], SortedDictPydanticAnnotation
]


class SortedListPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Returns pydantic_core.CoreSchema that defines how Pydantic should validate and
        serialize this class.

        - Validating from JSON: Validate as an iterable and pass to SortedList
          constructor
        - Validating from Python:
            - If it's already a SortedList, do nothing
            - If it's an iterable, pass to SortedList constructor
        - Serialization: Convert to a list
        """
        if cls is SortedListPydanticAnnotation:
            # Used as annotation, e.g., Annotated[..., SortedListPydanticAnnotation]
            try:
                cls = _get_constructor(source_type)
            except _UnsupportedSourceTypeError as e:
                msg = (
                    "Expected subclass of sortedcontainers.SortedList, "
                    f"got '{e.parsed}' parsed from annotation '{source_type}'."
                )
                raise UnsupportedSourceTypeError(msg) from e

        # Schema for when the input is already an instance of this class
        instance_schema = core_schema.is_instance_schema(cls)

        # Get schema for Iterable type based on source type has arguments
        args = get_args(source_type)
        if args:
            iterable_t_schema = handler.generate_schema(Iterable[args[0]])  # type: ignore[valid-type]
        else:
            iterable_t_schema = handler.generate_schema(Iterable)

        # Schema for when the input is an iterable
        from_iterable_schema = core_schema.no_info_after_validator_function(
            function=cls, schema=iterable_t_schema
        )

        # Union of the two schemas
        # Only include instance_schema if there are no type arguments
        # Otherwise an existing instance with wrong argument types won't be coerced
        python_schema: core_schema.CoreSchema
        if args:
            python_schema = from_iterable_schema
        else:
            python_schema = core_schema.union_schema([instance_schema, from_iterable_schema])

        # Serializer that converts an instance to a list
        as_list_serializer = core_schema.plain_serializer_function_ser_schema(list)

        return core_schema.json_or_python_schema(
            json_schema=from_iterable_schema,
            python_schema=python_schema,
            serialization=as_list_serializer,
        )


class SortedList(sortedcontainers.SortedList[_T], SortedListPydanticAnnotation):
    pass


class SortedKeyList(sortedcontainers.SortedKeyList[_T, _OrderableT], SortedList[_T]):
    pass


AnnotatedSortedList = Annotated[sortedcontainers.SortedList[_T], SortedListPydanticAnnotation]  # type: ignore[misc]

# Don't define AnnotatedSortedKeyList
# We generally don't expect people to directly use SortedKeyList in a field definition
# They should use SortedList instead


class SortedSetPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Returns pydantic_core.CoreSchema that defines how Pydantic should validate and
        serialize this class.

        - Validating from JSON: Validate as an iterable and pass to SortedSet
          constructor
        - Validating from Python:
            - If it's already a SortedSet, do nothing
            - If it's a set, parse as a set and pass to SortedSet constructor
            - If it's an iterable, pass to SortedSet constructor
        - Serialization: Convert to a list
        """
        if cls is SortedSetPydanticAnnotation:
            # Used as annotation, i.e., Annotated[..., SortedSetPydanticAnnotation]
            try:
                cls = _get_constructor(source_type)
            except _UnsupportedSourceTypeError as e:
                msg = (
                    "Expected subclass of sortedcontainers.SortedSet, "
                    f"got '{e.parsed}' parsed from annotation '{source_type}'."
                )
                raise UnsupportedSourceTypeError(msg) from e

        # Schema for when the input is already an instance of this class
        instance_schema = core_schema.is_instance_schema(cls)

        # Get schema for Iterable type based on source type has arguments
        args = get_args(source_type)
        if args:
            set_t_schema = handler.generate_schema(Set[args[0]])  # type: ignore[valid-type]
            iterable_t_schema = handler.generate_schema(Iterable[args[0]])  # type: ignore[valid-type]
        else:
            set_t_schema = handler.generate_schema(Set)
            iterable_t_schema = handler.generate_schema(Iterable)

        # Schema for when the input is a set
        from_set_schema = core_schema.no_info_after_validator_function(
            function=cls, schema=set_t_schema
        )

        # Schema for when the input is an iterable
        from_iterable_schema = core_schema.no_info_after_validator_function(
            function=cls, schema=iterable_t_schema
        )

        # Union of the schemas
        # Only include instance_schema if there are no type arguments
        # Otherwise an existing instance with wrong argument types won't be coerced
        if args:
            python_schema = core_schema.union_schema([from_set_schema, from_iterable_schema])
        else:
            python_schema = core_schema.union_schema(
                [instance_schema, from_set_schema, from_iterable_schema]
            )

        # Serializer that converts an instance to a list
        as_list_serializer = core_schema.plain_serializer_function_ser_schema(list)

        return core_schema.json_or_python_schema(
            json_schema=from_set_schema,
            python_schema=python_schema,
            serialization=as_list_serializer,
        )


class SortedSet(sortedcontainers.SortedSet[_HashableT], SortedSetPydanticAnnotation):
    pass


AnnotatedSortedSet = Annotated[sortedcontainers.SortedSet[_HashableT], SortedSetPydanticAnnotation]


@dataclass(frozen=True)
class Key:
    key: Callable[[Any], "SupportsRichComparison"]

    def __get_pydantic_core_schema__(
        self, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        try:
            constructor = _get_constructor(source_type)
        except _UnsupportedSourceTypeError as e:
            msg = (
                "Expected subclass of a sortedcontainers or sortedcontainers_pydantic class, "
                f"got '{e.parsed}' parsed from annotation '{source_type}'."
            )
            raise UnsupportedSourceTypeError(msg) from e
        # sortedcontainers.SortedList has magic behavior where the SortedList constructor returns
        # SortedKeyList if given a key. We match that behavior for our SortedList.
        if constructor is SortedList:
            constructor = SortedKeyList
        return core_schema.no_info_after_validator_function(
            function=partial(constructor, key=self.key),
            schema=handler(source_type),
        )



from typing import Tuple

from pydantic import BaseModel

from sortedcontainers_pydantic import SortedDict, SortedList, SortedSet


class MyModel(BaseModel):
    sorted_dict: SortedDict[str, int]
    sorted_list: SortedList[float]
    sorted_set: SortedSet[Tuple[int, int]]


m = MyModel(
    sorted_dict={"carol": 2, "alice": 3, "bob": 1},
    sorted_list=[3.0, 2.0, 1.0],
    sorted_set={(2, 9), (3, 5), (1, 3)},
)
print(repr(m))
#> MyModel(sorted_dict=SortedDict({'alpha': 3, 'bravo': 1, 'charlie': 2}), sorted_list=SortedList([1.0, 2.0, 3.0]), sorted_set=SortedSet([(1, 3), (2, 9), (3, 5)]))
print(m.model_dump_json())
#> {"sorted_dict":{"alpha":3,"bravo":1,"charlie":2},"sorted_list":[1.0,2.0,3.0],"sorted_set":[[1,3],[2,9],[3,5]]}

m2 = MyModel.model_validate_json(m.model_dump_json())
assert m == m2
print(repr(m2))
#> MyModel(sorted_dict=SortedDict({'alpha': 3, 'bravo': 1, 'charlie': 2}), sorted_list=SortedList([1.0, 2.0, 3.0]), sorted_set=SortedSet([(1, 3), (2, 9), (3, 5)]))
